/******************************************************************************
 * FIRESTARTER - A Processor Stress Test Utility 
 * Copyright (C) 2013 TU Dresden, Center for Information Services and High
 * Performance Computing
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contact: daniel.hackenberg@tu-dresden.de
 *****************************************************************************/

/* Initial values for computation */
#define KERNEL_INT_INIT 546743
#define KERNEL_FP_INIT 0.83475

/* Default Settings  */
#define FALSE 0
#define TRUE  1

/* current version */
#define VERSION_MAJOR 1
#define VERSION_MINOR 0
#define COPYRIGHT_YEAR 2013

#define _GNU_SOURCE
#include <sched.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <assert.h>
#include <signal.h>
#include <errno.h>

#include <ctype.h>
/*  
 * Header for local functions
 */
#include "work.h"
#include "cpu.h"

/*
 * 
 */
pid_t getpid(void);
unsigned int sleep(unsigned int seconds);

/*
 * set by evaluate_environment function
 */
long long BUFFERSIZEMEM;
int ALIGNMENT=64; 

unsigned int NUM_THREADS=0;
int BURST_LENGTH=0,FUNCTION=0,TIMEOUT=0;

long long INT_INIT[2]={1,-1};
double FP_INIT[2]={2.0,0.5};

char *error_msg=NULL;

/*
 * CPU bindings of threads
 */
cpu_set_t cpuset;
unsigned long long *cpu_bind;

/*
 * displays hardware information
 */
unsigned int verbose=0;

/*
 * data structure for hardware detection
 */
cpu_info_t *cpuinfo=NULL;

mydata_t* mdp;

/*
 * watchdog timer
 */
pthread_t watchdog;
typedef struct watchdog_args
{
 pid_t pid;
 int timeout;
} watchdog_arg_t;
watchdog_arg_t watchdog_arg;

/* stops watchdog thread if benchmark finishes before timeout*/
static void sigusr1_handler () {
 pthread_exit(0);
}

/* stops benchmark if timeout is reached*/
static void *watchdog_timer(void *arg)
{
  sigset_t  signal_mask; 
  
  sigemptyset (&signal_mask);
  sigaddset (&signal_mask, SIGINT);
  sigaddset (&signal_mask, SIGTERM);
  pthread_sigmask (SIG_BLOCK, &signal_mask, NULL);
  
  signal(SIGUSR1,sigusr1_handler);
  
  if (((watchdog_arg_t*)arg)->timeout>0)
  {
     sleep(((watchdog_arg_t*)arg)->timeout);
     kill(((watchdog_arg_t*)arg)->pid,SIGTERM);
  }
  pthread_exit(0);
}

void evaluate_environment();

void* init()
{
   unsigned int i;
   int t,j;
   unsigned long long tmp;

   cpu_set(cpu_bind[0]);
   mdp->cpuinfo=cpuinfo;

  BUFFERSIZEMEM=sizeof(char)*(2*mdp->buffersize[0]+mdp->buffersize[1]+mdp->buffersize[2]+mdp->rambuffersize+ALIGNMENT+2*sizeof(unsigned long long));

   if (BUFFERSIZEMEM <= 0)
   {
	fprintf( stderr, "Error: Determine BUFFERSIZEMEM failed\n"); fflush( stderr);
	exit( 127 );
   }
   mdp->INT_INIT=INT_INIT;
   mdp->FP_INIT=FP_INIT;
 
   
   if ((NUM_THREADS>mdp->cpuinfo->num_cores)||(NUM_THREADS==0)) NUM_THREADS=mdp->cpuinfo->num_cores;
   
   mdp->num_threads=NUM_THREADS;
   mdp->threads=_mm_malloc(NUM_THREADS*sizeof(pthread_t),ALIGNMENT);
   mdp->thread_comm=_mm_malloc(NUM_THREADS*sizeof(int),ALIGNMENT);
   if ((mdp->threads==NULL)||(mdp->thread_comm==NULL))
   {
     fprintf( stderr, "Error: Allocation of structure mydata_t failed (2)\n" ); fflush( stderr );
     exit( 127 );
   }

  mdp->threaddata = _mm_malloc(mdp->num_threads*sizeof(threaddata_t),ALIGNMENT);

  if (generic_num_packages()!=-1) mdp->threads_per_package=calloc(generic_num_packages(),sizeof(int));
  else mdp->threads_per_package=calloc(1,sizeof(int));

  for (i=0;i<NUM_THREADS;i++)
  {
    if (get_pkg(cpu_bind[i])!=-1) mdp->threaddata[i].package=get_pkg(cpu_bind[i]);
    else mdp->threaddata[i].package=0;
    mdp->threads_per_package[mdp->threaddata[i].package]++;
  }
  
  // create threads
  for (t=0;t<mdp->num_threads;t++)
  {
    mdp->threaddata[t].cpuinfo=(cpu_info_t*)_mm_malloc( sizeof( cpu_info_t ),ALIGNMENT);
    if ( mdp->cpuinfo == 0 )
    {
      fprintf( stderr, "Error: Allocation of structure mydata_t failed (3)\n" ); fflush( stderr );
      exit( 127 );
    }

    memcpy(mdp->threaddata[t].cpuinfo,mdp->cpuinfo,sizeof(cpu_info_t));
    mdp->ack=0;
    mdp->threaddata[t].thread_id=t;
    mdp->threaddata[t].cpu_id=cpu_bind[t];
    mdp->threaddata[t].data=mdp;
    mdp->thread_comm[t]=THREAD_INIT;
    for (i=0;i<3;i++){mdp->threaddata[t].buffersize[i]=mdp->buffersize[i];}
    mdp->threaddata[t].buffersizeMem=BUFFERSIZEMEM;
    mdp->threaddata[t].alignment=ALIGNMENT;
    mdp->threaddata[t].FUNCTION=FUNCTION;
    mdp->threaddata[t].BURST_LENGTH=BURST_LENGTH;
    if (t)
    {
     pthread_create(&(mdp->threads[t]),NULL,thread,(void*)(&(mdp->threaddata[t])));
     while (!mdp->ack);
    }
  }
  mdp->ack=0;mdp->done=0;
  
  cpu_set(cpu_bind[0]);

   //Memory-Allocation Master-Thread
  switch(FUNCTION)
  {
	case 0:
	case 1:
	case 2:
	case 3:
	case 4:
	case 5:
		mdp->bufferMem = _mm_malloc( BUFFERSIZEMEM,ALIGNMENT );
		if ((mdp->bufferMem == 0)||(mdp->bufferMem == (void*) -1ULL))
   		{
      			fprintf( stderr, "Error: Allocation of Memory failed\n" ); fflush( stderr );
      			exit( 127 );
   		}
		break;
	default:
    		fprintf( stderr, "Error: Allocation of buffer failed: unknown mode:%i \n",FUNCTION );      
    		exit( 1 );
  }

   switch(FUNCTION){
   case 0:
   case 1:
   case 2:
   case 4:
   case 5:

	break;
   case 3:
	// Fill Cache-/Memory-buffer   
   	tmp=32*BURST_LENGTH*sizeof(double);
	// Fill RAM
        for (i=0;i<=BUFFERSIZEMEM-tmp;i+=tmp)
        {
		for(j=0;j<4*BURST_LENGTH;j++)
		  *((unsigned long long*)((unsigned long long)mdp->bufferMem+i+j*sizeof(unsigned long long)))=18446744073709551615llu; //1111...
        	  //*((double*)((unsigned long long)mdp->bufferMem+i+j*sizeof(double)))=FP_INIT[0];
      		for(j=4*BURST_LENGTH;j<8*BURST_LENGTH;j++)
        	  *((double*)((unsigned long long)mdp->bufferMem+i+j*sizeof(double)))=-1.0*FP_INIT[1];
      		for(j=8*BURST_LENGTH;j<12*BURST_LENGTH;j++)
       		  *((double*)((unsigned long long)mdp->bufferMem+i+j*sizeof(double)))=FP_INIT[0];
    		for(j=12*BURST_LENGTH;j<16*BURST_LENGTH;j++)
        	  *((double*)((unsigned long long)mdp->bufferMem+i+j*sizeof(double)))=-1.0*FP_INIT[1];
      		for(j=16*BURST_LENGTH;j<20*BURST_LENGTH;j++)
        	  *((double*)((unsigned long long)mdp->bufferMem+i+j*sizeof(double)))=-1.0*FP_INIT[0];
      		for(j=20*BURST_LENGTH;j<24*BURST_LENGTH;j++)
        	  *((double*)((unsigned long long)mdp->bufferMem+i+j*sizeof(double)))=FP_INIT[1];
      		for(j=24*BURST_LENGTH;j<28*BURST_LENGTH;j++)
        	  *((double*)((unsigned long long)mdp->bufferMem+i+j*sizeof(double)))=-1.0*FP_INIT[0];
      		for(j=28*BURST_LENGTH;j<32*BURST_LENGTH;j++)
        	  *((double*)((unsigned long long)mdp->bufferMem+i+j*sizeof(double)))=FP_INIT[1];
   	}
   	break;	
    default:
    fprintf( stderr, "Error: initialization failed: unknown mode:%i \n",FUNCTION );      
    exit( 1 );
  }


  printf("  wait for threads memory initialization \n");fflush(stdout);
  /* wait for threads to finish their initialization*/  
  for (t=1;t<mdp->num_threads;t++)
  {
     mdp->ack=0;
     mdp->thread_comm[t]=2;
     while (!mdp->ack);
  }
  mdp->ack=0;
  printf("    ...done\n");fflush(stdout);
  if ((num_packages()!=-1)&&(num_cores_per_package()!=-1)&&(num_threads_per_core()!=-1))  printf("  num_packages: %i, %i cores per package, %i threads per core\n",num_packages(),num_cores_per_package(),num_threads_per_core());
  printf("  using %i threads\n",NUM_THREADS);fflush(stdout);
  for (i=0;i<NUM_THREADS;i++) if ((get_pkg(cpu_bind[i])!=-1)&&(get_core_id(cpu_bind[i])!=-1)) printf("    - Thread %i runs on CPU %llu, core %i in package: %i\n",i,cpu_bind[i],get_core_id(cpu_bind[i]),get_pkg(cpu_bind[i]));fflush(stdout);

  //start watchdog thread
  watchdog_arg.pid=getpid();
  watchdog_arg.timeout=TIMEOUT;
  pthread_create(&watchdog,NULL,watchdog_timer,&watchdog_arg);
  
  return (void*)mdp;
}
/********************************************************************/
/*************** End of interface implementations *******************/
/********************************************************************/

/* Reads the environment variables used by this kernel. */
void evaluate_environment( )
{
   unsigned int i,j = 0;
   char arch[16];
   int errors = 0;
  
   cpuinfo=(cpu_info_t*)_mm_malloc( sizeof( cpu_info_t ),64);
   init_cpuinfo(cpuinfo,verbose);

   error_msg=malloc(256);  
   
   CPU_ZERO(&cpuset);

   //use all CPUs if not defined otherwise
   for (i=0;i<cpuinfo->num_cores;i++) {if (cpu_allowed(i)) {CPU_SET(i,&cpuset);NUM_THREADS++;}}

   if (NUM_THREADS==0) {errors++;sprintf(error_msg,"Could not determine number of CPUs!");}
   else
   {
     cpu_bind=(unsigned long long*)malloc((NUM_THREADS+1)*sizeof(unsigned long long));
     for(i=0;i<cpuinfo->num_cores;i++)
     {
      if (CPU_ISSET(i,&cpuset)) {cpu_bind[j]=i;j++;}
     }
   }
   
      INT_INIT[0]=KERNEL_INT_INIT;
      INT_INIT[1]=INT_INIT[0]*-1;
   
      FP_INIT[0]=KERNEL_FP_INIT;
      FP_INIT[1]=(1.0/FP_INIT[0]);
   

   mdp = (mydata_t*)_mm_malloc( sizeof( mydata_t ),ALIGNMENT);
   if ( mdp == 0 )
   {
      fprintf( stderr, "Error: Allocation of structure mydata_t failed (1)\n" ); fflush( stderr );
      exit( 127 );
   }
   memset(mdp,0,sizeof(mydata_t));

   get_architecture(arch,sizeof(arch));
   if (strcmp(arch,"x86_64")) {
      fprintf( stderr, "Error: wrong architecture: %s, x86_64 required \n",arch );
      exit( 1 );
   }
    
   if (!feature_available("SSE2")) {
        fprintf( stderr, "Error: SSE2 not supported!\n" );
        exit( 1 );
   }

if (strcmp("GenuineIntel",cpuinfo->vendor) == 0){
  switch(cpuinfo->model){
	case 28:
	case 38:
	case 54:
		fprintf( stderr, "Error: Intel Atom Processor is not supported!\n" );
 		exit( 1 );
		break;
	case 3:
	case 4:
	case 6:
	case 13:
	case 15:
	case 22:
	case 23:
	case 29:
		fprintf( stderr, "Error: Your Intel Processor is not supported!\n" );
 		exit( 1 );
		break;
	default:
		break;
	}

  if (feature_available("SSE2")){
	ALIGNMENT=128;
	switch(cpuinfo->model){
	case 26:
	case 30:
	case 37:
	case 44:
	case 46:	//Nehalem-EX
	case 47:	//Westmere-EX
		FUNCTION=0;
		break;
	}
  }
  if (feature_available("SSE2") && feature_available("SMT")){
	ALIGNMENT=128;
	switch(cpuinfo->model){
	  case 26:
	  case 30:
	  case 37:
	  case 44:
	  case 46:	//Nehalem-EX
	  case 47:	//Westmere-EX
		FUNCTION=1;
		break;
	}
  }
 
  if (feature_available("AVX")){
	ALIGNMENT=256;
	switch(cpuinfo->model){
	  case 42:
	  case 58:
		FUNCTION=2;
		break;
	  case 45:		//Sandy-Bridge-E{N|P}
		FUNCTION=4;
		break;
	}
  }
  if (feature_available("AVX") && feature_available("SMT")){
	ALIGNMENT=256;
	switch(cpuinfo->model){
	  case 42:
	  case 58:
		FUNCTION=3;
		break;
	  case 45:		//Sandy-Bridge-E{N|P}
		FUNCTION=5;
		break;
	}
 }
}
else if (strcmp("AuthenticAMD",cpuinfo->vendor) == 0){
	fprintf( stderr, "Error: AMD CPUs are not supported yet!\n" );
 	exit( 1 );	
}
else{
	fprintf( stderr, "Error: CPU not supported!\n" );
 	exit( 1 );
}


   switch(FUNCTION){
     case 0: printf("\n  Take SSE2-PATH (Nehalem,Westmere)\n");
	     break;
     case 1: printf("\n  Take SSE2-SMT-PATH (Nehalem,Westmere)\n");
	     break;
     case 2: printf("\n  Take AVX-PATH (Sandy-Bridge,Ivy-Bridge)\n");
	     break;
     case 3: printf("\n  Take AVX-SMT-PATH (Sandy-Bridge,Ivy-Bridge)\n");
	     break;
     case 4: printf("\n  Take AVX-PATH (Sandy-Bridge-E{N|P})\n");
	     break;
     case 5: printf("\n  Take AVX-SMT-PATH (Sandy-Bridge-E{N|P})\n");
	     break;
     default: fprintf( stderr, "Error: CPU not supported!\n" );
 	      exit( 1 );
   }

   /* Determine Cachesizes */
   for(i=0;i<3;i++) mdp->buffersize[i] = 0;
   if(cpuinfo->Cachelevels)
   {
	printf("  Used buffersizes per physical core:\n");
   	for(i=0;i<cpuinfo->Cachelevels;i++)
        {
        	if (cpuinfo->Cache_unified[i])
		{
		 mdp->buffersize[i] = cpuinfo->U_Cache_Size[i];
		}
        	else{
		 mdp->buffersize[i] = cpuinfo->D_Cache_Size[i];
		}
        	if (i==2){
			mdp->buffersize[i] = mdp->buffersize[i] / (cpuinfo->num_cores_per_package);
		}
		printf("    - L%d-Cache: %d Bytes\n",i+1,mdp->buffersize[i]);
     }
   } 
   /* Determine RAMsize per Core */   
   mdp->rambuffersize = 268435456; //256 MiByte 
   printf("    - Memory: %llu Bytes\n\n",mdp->rambuffersize);
}

void show_help(void) {
	printf("FIRESTARTER - A Processor Stress Test Utility, Version %i.%i\n",VERSION_MAJOR,VERSION_MINOR);
	printf("Copyright (C) %i TU Dresden, Center for Information Services and High Performance Computing\n",COPYRIGHT_YEAR);
	printf("This program comes with ABSOLUTELY NO WARRANTY; for details run `FIRESTARTER -w'.\n");
	printf("This is free software, and you are welcome to redistribute it\nunder certain conditions; run `FIRESTARTER -c' for details.\n");

   printf("\nUsage: FIRESTARTER [options] TIME (sec.)\n"\
          "\nOptions:\n\n"\
          " -h              display usage information\n"\
          " -v              display version information\n"\
          " -i              display detailed hardware information\n"\
          " -w              display warranty information\n"\
          " -c              display copyright information\n\n");
}

void show_warranty(void) {
	printf("FIRESTARTER - A Processor Stress Test Utility, Version %i.%i\n",VERSION_MAJOR,VERSION_MINOR);
	printf("Copyright (C) %i TU Dresden, Center for Information Services and High Performance Computing\n",COPYRIGHT_YEAR);
	printf("\nThis program is distributed in the hope that it will be useful,\nbut WITHOUT ANY WARRANTY; without even the implied warranty of\nMERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\nGNU General Public License for more details.\n\n");
	printf("You should have received a copy of the GNU General Public License\nalong with this program.  If not, see <http://www.gnu.org/licenses/>.\n\n");
}

void show_copyright(void) {
	printf("FIRESTARTER - A Processor Stress Test Utility, Version %i.%i\n",VERSION_MAJOR,VERSION_MINOR);
	printf("Copyright (C) %i TU Dresden, Center for Information Services and High Performance Computing\n",COPYRIGHT_YEAR);
	printf("\nThis program is free software: you can redistribute it and/or modify\nit under the terms of the GNU General Public License as published by\nthe Free Software Foundation, either version 3 of the License, or\n(at your option) any later version.\n\n");
	printf("You should have received a copy of the GNU General Public License\nalong with this program.  If not, see <http://www.gnu.org/licenses/>.\n\n");
}

void show_version(void) {
	printf("FIRESTARTER - A Processor Stress Test Utility, Version %i.%i\n",VERSION_MAJOR,VERSION_MINOR);
	printf("Copyright (C) %i TU Dresden, Center for Information Services and High Performance Computing\n",COPYRIGHT_YEAR);
}

int getopt(char *argument, char *option) {
   if( argument[0]=='-' && argument[1]==option[0] )
      return TRUE;
   return FALSE;
}

int main(int argc, char *argv[]){

	if (argc == 2){
		if (getopt(argv[1],"h") == TRUE){show_help();return EXIT_SUCCESS;}
		else if (getopt(argv[1],"v") == TRUE){show_version();return EXIT_SUCCESS;}
		else if (getopt(argv[1],"w") == TRUE){show_warranty();return EXIT_SUCCESS;}
		else if (getopt(argv[1],"c") == TRUE){show_copyright();return EXIT_SUCCESS;}
		else if (getopt(argv[1],"i") == TRUE){
		  printf("\nUsage: FIRESTARTER -i TIME (sec.)\n");
		  return EXIT_FAILURE;
   		}
		else{
		  TIMEOUT=atoi(argv[1]);
		}
	}
	else if (argc==3){
		if (getopt(argv[1],"i") == TRUE){verbose=1;TIMEOUT=atoi(argv[2]);}
		else if (getopt(argv[2],"i") == TRUE){verbose=1;TIMEOUT=atoi(argv[1]);}
		else{show_help();return EXIT_FAILURE;}
	}
	else{
		show_help();
		return EXIT_FAILURE;
	}

	printf("FIRESTARTER - A Processor Stress Test Utility, Version %i.%i\n",VERSION_MAJOR,VERSION_MINOR);
	printf("Copyright (C) %i TU Dresden, Center for Information Services and High Performance Computing\n",COPYRIGHT_YEAR);
	printf("This program comes with ABSOLUTELY NO WARRANTY; for details run `FIRESTARTER -w'.\n");
	printf("This is free software, and you are welcome to redistribute it\nunder certain conditions; run `FIRESTARTER -c' for details.\n");

	evaluate_environment();
	if (FUNCTION == 0){BURST_LENGTH = 4;}
	else{BURST_LENGTH = 8;}
        init();
	_work(mdp->rambuffersize,FUNCTION,mdp);

	return EXIT_SUCCESS;
}
